/** \file     my_io_dimacs.h
 *  \brief    IO functions for data files in DIMACS CNF format
 *  \author   Takahisa Toda
 *  \note     See http://www.cs.ubc.ca/~hoos/SATLIB/benchm.html for benchmark problems and details of DIMACS CNF format.
 */
#ifndef MY_IO_DIMACS_H
#define MY_IO_DIMACS_H

#include <stdio.h>
#include <malloc.h>
#include "my_def.h"

#define CH_DIMACS_COMMENT   'c' //!< comment line character of DIMACS format
#define CH_DIMACS_PROBLEM   'p' //!< problem line character of DIMACS format
#define MAX_LINELEN         (UINTMAX_C(1) << 20)  //!< the maximum length of a single line in a datafile

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/** \brief  An explicite data represention of a set family
 *  \note
 *  - Let sf point to a setfam data structure.
 *  - sf->elem[i-1][j-1] corresponds to the j-th element in the i-th set. 
 *  - Each set ends with INFTY, that is,
 *  - sf->elem[i-1][k] == INFTY if the i-th set contains exactly k elements.
 */
struct setfam {
  uintmax_t card;      //!< the number of sets with repetition, i.e. cardinality
  uintmax_t memlen;    //!< the length of mem
  itemval   **elem;    //!< double pointer to an element in mem
  itemval   *mem;      //!< memory area (array) of elements, which are partitioned into sets using INFTY as delimiter.
};

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/
extern int              getfileinfo_dimacs          (uintmax_t *pn, itemval *pm, uintmax_t *pt, FILE *in);
extern struct setfam    *create_setfam_from_dimacs  (uintmax_t n, uintmax_t t, FILE *in);
extern void             destroy_setfam              (struct setfam *sf);


/** \brief Compute a dynamically allocated memory size in byte.
 *  \param h      Pointer to a setfam data structure
 *  \return The computed size in byte
 */
static inline uintmax_t setfam_allocsize            (struct setfam *sf)
{
  return malloc_usable_size(sf) + malloc_usable_size(sf->elem) + malloc_usable_size(sf->mem); 
}

#endif /*MY_IO_DIMACS_H*/
